from __future__ import annotations

from enum import Enum, auto
from typing import Optional, List, Iterator

from .capability_interfaces import ParameterIF, CapabilityStepIF, CapabilitySequenceIF, \
    CapabilityRequestIF
from .product_interfaces import FutureProductIF


class CapabilityStep(CapabilityStepIF):
    """
    Class that represents a step in a capability sequence. A step is of a certain type and
    certain step types can have a value
    """
    def __init__(self, step_type: CapabilityStepType, step_value: Optional[str]):
        self.step_type = step_type
        self.step_value = step_value

    def __str__(self):
        if self.step_value:
            return f'{self.step_type.name} {self.step_value}'
        else:
            return f'{self.step_type.name}'


class CapabilitySequence(CapabilitySequenceIF):
    """
    Class that represents a capability sequence with a list of steps
    """
    def __init__(self, steps: List[CapabilityStep]):
        self.steps = steps

    def __iter__(self) -> Iterator:
        """
        Allows CapabilitySequence objects to be iterated over
        :return: Iterator object
        """
        return iter(self.steps)

    def __len__(self) -> int:
        """
        Allows len() to be called on CapabilitySequence objects
        :return: Number of capability steps
        """
        return len(self.steps)

    def __str__(self) -> str:
        """
        Provides a string representation of the object
        :return: String representation
        """
        str_steps = [str(step) for step in self]
        return ','.join(str_steps)


class Parameter(ParameterIF):
    pass


class FutureProduct(FutureProductIF):
    pass


class CapabilityStepType(Enum):
    """
    Enum that specifies the types of CapabilitySteps that are possible
    """
    PrepareAndRunWorkflow = 0
    AwaitQA = auto()
    AwaitWorkflow = auto()
    AwaitProduct = auto()
    AwaitParameter = auto()
    AwaitLargeAllocApproval = auto()

    @classmethod
    def from_string(cls, string: str) -> CapabilityStepType:
        strings = {
            'prepare-and-run-workflow': cls.PrepareAndRunWorkflow,
            'await-qa': cls.AwaitQA,
            'await-workflow': cls.AwaitWorkflow,
            'await-product': cls.AwaitProduct,
            'await-parameter': cls.AwaitParameter,
            'await-large-alloc-approval': cls.AwaitLargeAllocApproval
        }
        return strings[string]


class RequestState(Enum):
    """
    Enum that specifies the states that a capability request can be in
    """
    Complete = 0
    Executing = 1
    Ready = 2
    Cancelled = 3


class ExecutionState(Enum):
    """
    Enum that specifies the states that a capability execution can find itself in
    """
    Complete = 0
    ExecutingStep = 1
    Ready = 2
    Cancelled = 3
    Failed = -1


class ExecutionPriority(Enum):
    """
    Enum that specifies the priority of a particular capability execution
    """
    High = 25
    Default = 50
    Low = 100
